@@ -2,15 +2,19 @@
 Connection Pooling
 
 """
+import datetime
 import logging
+import os
 import threading
 import time
 import weakref
 
+import psycopg2
+
 LOGGER = logging.getLogger(__name__)
 
 DEFAULT_IDLE_TTL = 60
-DEFAULT_MAX_SIZE = 1
+DEFAULT_MAX_SIZE = int(os.environ.get('QUERIES_MAX_POOL_SIZE', 1))
 
 
 class Connection(object):
@@ -23,6 +27,8 @@ class Connection(object):
     def __init__(self, handle):
         self.handle = handle
         self.used_by = None
+        self.executions = 0
+        self.exceptions = 0
 
     def close(self):
         """Close the connection
@@ -31,10 +37,14 @@ class Connection(object):
 
         """
         LOGGER.debug('Connection %s closing', self.id)
-        if self.busy:
+        if self.busy and not self.closed:
             raise ConnectionBusyError(self)
         with self._lock:
-            self.handle.close()
+            if not self.handle.closed:
+                try:
+                    self.handle.close()
+                except psycopg2.InterfaceError as error:
+                    LOGGER.error('Error closing socket: %s', error)
 
     @property
     def closed(self):
@@ -43,7 +53,7 @@ class Connection(object):
         :rtype: bool
 
         """
-        return self.handle.closed
+        return self.handle.closed != 0
 
     @property
     def busy(self):
@@ -125,11 +135,13 @@ class Pool(object):
     def __init__(self,
                  pool_id,
                  idle_ttl=DEFAULT_IDLE_TTL,
-                 max_size=DEFAULT_MAX_SIZE):
+                 max_size=DEFAULT_MAX_SIZE,
+                 time_method=None):
         self.connections = {}
         self._id = pool_id
         self.idle_ttl = idle_ttl
         self.max_size = max_size
+        self.time_method = time_method or time.time
 
     def __contains__(self, connection):
         """Return True if the pool contains the connection"""
@@ -151,11 +163,27 @@ class Pool(object):
             raise ValueError('Connection already exists in pool')
 
         if len(self.connections) == self.max_size:
+            LOGGER.warning('Race condition found when adding new connection')
+            try:
+                connection.close()
+            except (psycopg2.Error, psycopg2.Warning) as error:
+                LOGGER.error('Error closing the conn that cant be used: %s',
+                             error)
             raise PoolFullError(self)
         with self._lock:
             self.connections[id(connection)] = Connection(connection)
         LOGGER.debug('Pool %s added connection %s', self.id, id(connection))
 
+    @property
+    def busy_connections(self):
+        """Return a list of active/busy connections
+
+        :rtype: list
+
+        """
+        return [c for c in self.connections.values()
+                if c.busy and not c.closed]
+
     def clean(self):
         """Clean the pool by removing any closed connections and if the pool's
         idle has exceeded its idle TTL, remove all connections.
@@ -178,6 +206,34 @@ class Pool(object):
             self.remove(self.connections[cid].handle)
         LOGGER.debug('Pool %s closed', self.id)
 
+    @property
+    def closed_connections(self):
+        """Return a list of closed connections
+
+        :rtype: list
+
+        """
+        return [c for c in self.connections.values() if c.closed]
+
+    def connection_handle(self, connection):
+        """Return a connection object for the given psycopg2 connection
+
+        :param connection: The connection to return a parent for
+        :type connection: psycopg2.extensions.connection
+        :rtype: Connection
+
+        """
+        return self.connections[id(connection)]
+
+    @property
+    def executing_connections(self):
+        """Return a list of connections actively executing queries
+
+        :rtype: list
+
+        """
+        return [c for c in self.connections.values() if c.executing]
+
     def free(self, connection):
         """Free the connection from use by the session that was using it.
 
@@ -188,13 +244,13 @@ class Pool(object):
         """
         LOGGER.debug('Pool %s freeing connection %s', self.id, id(connection))
         try:
-            self._connection(connection).free()
+            self.connection_handle(connection).free()
         except KeyError:
             raise ConnectionNotFoundError(self.id, id(connection))
 
         if self.idle_connections == list(self.connections.values()):
             with self._lock:
-                self.idle_start = time.time()
+                self.idle_start = self.time_method()
         LOGGER.debug('Pool %s freed connection %s', self.id, id(connection))
 
     def get(self, session):
@@ -231,9 +287,8 @@ class Pool(object):
         :rtype: list
 
         """
-        return [self.connections[k] for k in self.connections if
-                not self.connections[k].busy and
-                not self.connections[k].closed]
+        return [c for c in self.connections.values()
+                if not c.busy and not c.closed]
 
     @property
     def idle_duration(self):
@@ -245,7 +300,7 @@ class Pool(object):
         """
         if self.idle_start is None:
             return 0
-        return time.time() - self.idle_start
+        return self.time_method() - self.idle_start
 
     @property
     def is_full(self):
@@ -266,7 +321,7 @@ class Pool(object):
         """
         cid = id(connection)
         try:
-            self._connection(connection).lock(session)
+            self.connection_handle(connection).lock(session)
         except KeyError:
             raise ConnectionNotFoundError(self.id, cid)
         else:
@@ -275,6 +330,15 @@ class Pool(object):
                     self.idle_start = None
         LOGGER.debug('Pool %s locked connection %s', self.id, cid)
 
+    @property
+    def locked_connections(self):
+        """Return a list of all locked connections
+
+        :rtype: list
+
+        """
+        return [c for c in self.connections.values() if c.locked]
+
     def remove(self, connection):
         """Remove the connection from the pool
 
@@ -287,11 +351,37 @@ class Pool(object):
         cid = id(connection)
         if cid not in self.connections:
             raise ConnectionNotFoundError(self.id, cid)
-        self._connection(connection).close()
+        self.connection_handle(connection).close()
         with self._lock:
             del self.connections[cid]
         LOGGER.debug('Pool %s removed connection %s', self.id, cid)
 
+    def report(self):
+        """Return a report about the pool state and configuration.
+
+        :rtype: dict
+
+        """
+        return {
+            'connections': {
+                'busy': len(self.busy_connections),
+                'closed': len(self.closed_connections),
+                'executing': len(self.executing_connections),
+                'idle': len(self.idle_connections),
+                'locked': len(self.busy_connections)
+            },
+            'exceptions': sum([c.exceptions
+                               for c in self.connections.values()]),
+            'executions': sum([c.executions
+                               for c in self.connections.values()]),
+            'full': self.is_full,
+            'idle': {
+                'duration': self.idle_duration,
+                'ttl': self.idle_ttl
+            },
+            'max_size': self.max_size
+        }
+
     def shutdown(self):
         """Forcefully shutdown the entire pool, closing all non-executing
         connections.
@@ -326,16 +416,6 @@ class Pool(object):
         with self._lock:
             self.max_size = size
 
-    def _connection(self, connection):
-        """Return a connection object for the given psycopg2 connection
-
-        :param connection: The connection to return a parent for
-        :type connection: psycopg2.extensions.connection
-        :rtype: Connection
-
-        """
-        return self.connections[id(connection)]
-
 
 class PoolManager(object):
     """The connection pool object implements behavior around connections and
@@ -392,16 +472,17 @@ class PoolManager(object):
 
         """
         with cls._lock:
-            cls._ensure_pool_exists(pid)
+            try:
+                cls._ensure_pool_exists(pid)
+            except KeyError:
+                LOGGER.debug('Pool clean invoked against missing pool %s', pid)
+                return
             cls._pools[pid].clean()
-
-            # If the pool has no open connections, remove it
-            if not len(cls._pools[pid]):
-                del cls._pools[pid]
+            cls._maybe_remove_pool(pid)
 
     @classmethod
     def create(cls, pid, idle_ttl=DEFAULT_IDLE_TTL, max_size=DEFAULT_MAX_SIZE,
-               pool_type=None):
+               time_method=None):
         """Create a new pool, with the ability to pass in values to override
         the default idle TTL and the default maximum size.
 
@@ -411,22 +492,33 @@ class PoolManager(object):
         A pool's max size defines the maximum number of connections that can
         be added to the pool to prevent unbounded open connections.
 
-        The pool_type value allows the invoking class to specify a different
-        type of Pool, allowing for different pooling behaviors.
-
         :param str pid: The pool ID
         :param int idle_ttl: Time in seconds for the idle TTL
         :param int max_size: The maximum pool size
-        :param class pool_type: The pool type to create
+        :param callable time_method: Override the use of :py:meth:`time.time`
+            method for time values.
         :raises: KeyError
 
         """
         if pid in cls._pools:
             raise KeyError('Pool %s already exists' % pid)
-        pool_class = pool_type or Pool
         with cls._lock:
             LOGGER.debug("Creating Pool: %s (%i/%i)", pid, idle_ttl, max_size)
-            cls._pools[pid] = pool_class(pid, idle_ttl, max_size)
+            cls._pools[pid] = Pool(pid, idle_ttl, max_size, time_method)
+
+    @classmethod
+    def free(cls, pid, connection):
+        """Free a connection that was locked by a session
+
+        :param str pid: The pool ID
+        :param connection: The connection to remove
+        :type connection: psycopg2.extensions.connection
+
+        """
+        with cls._lock:
+            LOGGER.debug('Freeing %s from pool %s', id(connection), pid)
+            cls._ensure_pool_exists(pid)
+            cls._pools[pid].free(connection)
 
     @classmethod
     def get(cls, pid, session):
@@ -443,18 +535,18 @@ class PoolManager(object):
             return cls._pools[pid].get(session)
 
     @classmethod
-    def free(cls, pid, connection):
-        """Free a connection that was locked by a session
+    def get_connection(cls, pid, connection):
+        """Return the specified :class:`~queries.pool.Connection` from the
+        pool.
 
         :param str pid: The pool ID
-        :param connection: The connection to remove
+        :param connection: The connection to return for
         :type connection: psycopg2.extensions.connection
+        :rtype: queries.pool.Connection
 
         """
         with cls._lock:
-            LOGGER.debug('Freeing %s from pool %s', id(connection), pid)
-            cls._ensure_pool_exists(pid)
-            cls._pools[pid].free(connection)
+            return cls._pools[pid].connection_handle(connection)
 
     @classmethod
     def has_connection(cls, pid, connection):
@@ -522,13 +614,11 @@ class PoolManager(object):
 
     @classmethod
     def remove_connection(cls, pid, connection):
-        """Remove a connection from the pool, closing it if is open. Will raise
-        an ActiveConnectionError if the connection is active prior to removal.
+        """Remove a connection from the pool, closing it if is open.
 
         :param str pid: The pool ID
         :param connection: The connection to remove
         :type connection: psycopg2.extensions.connection
-        :raises: ActiveConnectionError
         :raises: ConnectionNotFoundError
 
         """
@@ -539,7 +629,7 @@ class PoolManager(object):
     def set_idle_ttl(cls, pid, ttl):
         """Set the idle TTL for a pool, after which it will be destroyed.
 
-        :param str pid: The process id
+        :param str pid: The pool id
         :param int ttl: The TTL for an idle pool
 
         """
@@ -578,6 +668,19 @@ class PoolManager(object):
             cls._ensure_pool_exists(pid)
             return len(cls._pools[pid])
 
+    @classmethod
+    def report(cls):
+        """Return the state of the all of the registered pools.
+
+        :rtype: dict
+
+        """
+        return {
+            'timestamp': datetime.datetime.utcnow().isoformat(),
+            'process': os.getpid(),
+            'pools': dict([(i, p.report()) for i, p in cls._pools.items()])
+        }
+
     @classmethod
     def _ensure_pool_exists(cls, pid):
         """Raise an exception if the pool has yet to be created or has been
@@ -590,64 +693,63 @@ class PoolManager(object):
         if pid not in cls._pools:
             raise KeyError('Pool %s has not been created' % pid)
 
+    @classmethod
+    def _maybe_remove_pool(cls, pid):
+        """If the pool has no open connections, remove it
 
-class ActiveConnectionError(Exception):
-    """Raised when removing an active connection from a pool"""
+        :param str pid: The pool id to clean
 
-    def __init__(self, pid, connection):
-        self.pid = pid
-        self.cid = connection.id
-
-    def __str__(self):
-        return 'Connection %s in pool %s is active' % (self.cid, self.pid)
+        """
+        if not len(cls._pools[pid]):
+            del cls._pools[pid]
 
 
-class ActivePoolError(Exception):
-    """Raised when removing a pool that has active connections"""
-
-    def __init__(self, pid):
-        self.pid = pid
-
-    def __str__(self):
-        return 'Pool %s has at least one active connection' % self.pid
+class QueriesException(Exception):
+    """Base Exception for all other Queries exceptions"""
+    pass
 
 
-class ConnectionBusyError(Exception):
-    """Raised when trying to lock a connection that is already busy"""
-
+class ConnectionException(QueriesException):
     def __init__(self, cid):
         self.cid = cid
 
-    def __str__(self):
-        return 'Connection %s is busy' % self.cid
+
+class PoolException(QueriesException):
+    def __init__(self, pid):
+        self.pid = pid
 
 
-class ConnectionNotFoundError(Exception):
-    """Raised if a specific connection is not found in the pool"""
-
+class PoolConnectionException(PoolException):
     def __init__(self, pid, cid):
         self.pid = pid
         self.cid = cid
 
+
+class ActivePoolError(PoolException):
+    """Raised when removing a pool that has active connections"""
+    def __str__(self):
+        return 'Pool %s has at least one active connection' % self.pid
+
+
+class ConnectionBusyError(ConnectionException):
+    """Raised when trying to lock a connection that is already busy"""
+    def __str__(self):
+        return 'Connection %s is busy' % self.cid
+
+
+class ConnectionNotFoundError(PoolConnectionException):
+    """Raised if a specific connection is not found in the pool"""
     def __str__(self):
         return 'Connection %s not found in pool %s' % (self.cid, self.pid)
 
 
-class NoIdleConnectionsError(Exception):
+class NoIdleConnectionsError(PoolException):
     """Raised if a pool does not have any idle, open connections"""
-
-    def __init__(self, pid):
-        self.pid = pid
-
     def __str__(self):
         return 'Pool %s has no idle connections' % self.pid
 
 
-class PoolFullError(Exception):
+class PoolFullError(PoolException):
     """Raised when adding a connection to a pool that has hit max-size"""
-
-    def __init__(self, pid):
-        self.pid = pid
-
     def __str__(self):
         return 'Pool %s is at its maximum capacity' % self.pid
