@@ -23,23 +23,14 @@ Example Use:
 
 """
 import logging
+import socket
+import warnings
 
-from psycopg2 import extensions
-from psycopg2 import extras
-
-from tornado import concurrent
-from tornado import gen
-from tornado import ioloop
-from tornado import stack_context
+from tornado import concurrent, ioloop
+from psycopg2 import extras, extensions
 import psycopg2
 
-from queries import pool
-from queries import results
-from queries import session
-from queries import utils
-
-from queries import DEFAULT_URI
-from queries import PYPY
+from queries import pool, results, session, utils
 
 LOGGER = logging.getLogger(__name__)
 
@@ -107,8 +98,8 @@ class Results(results.Results):
         self.cursor = cursor
         self._cleanup = cleanup
         self._fd = fd
+        self._freed = False
 
-    @gen.coroutine
     def free(self):
         """Release the results and connection lock from the TornadoSession
         object. This **must** be called after you finish processing the results
@@ -118,7 +109,13 @@ class Results(results.Results):
         requests.
 
         """
-        yield self._cleanup(self.cursor, self._fd)
+        self._freed = True
+        self._cleanup(self.cursor, self._fd)
+
+    def __del__(self):
+        if not self._freed:
+            LOGGER.warning('Auto-freeing result on deletion')
+            self.free()
 
 
 class TornadoSession(session.Session):
@@ -139,7 +136,7 @@ class TornadoSession(session.Session):
     :param int pool_max_size: The maximum size of the pool to use
 
     """
-    def __init__(self, uri=DEFAULT_URI,
+    def __init__(self, uri=session.DEFAULT_URI,
                  cursor_factory=extras.RealDictCursor,
                  pool_idle_ttl=pool.DEFAULT_IDLE_TTL,
                  pool_max_size=DEFAULT_MAX_POOL_SIZE,
@@ -155,16 +152,21 @@ class TornadoSession(session.Session):
 
         """
         self._connections = dict()
-        self._futures = dict()
-
+        self._cleanup_callback = None
         self._cursor_factory = cursor_factory
-        self._ioloop = io_loop or ioloop.IOLoop.instance()
+        self._futures = dict()
+        self._ioloop = io_loop or ioloop.IOLoop.current()
         self._pool_manager = pool.PoolManager.instance()
+        self._pool_max_size = pool_max_size
+        self._pool_idle_ttl = pool_idle_ttl
         self._uri = uri
+        self._ensure_pool_exists()
 
-        # Ensure the pool exists in the pool manager
+    def _ensure_pool_exists(self):
+        """Create the pool in the pool manager if it does not exist."""
         if self.pid not in self._pool_manager:
-            self._pool_manager.create(self.pid, pool_idle_ttl, pool_max_size)
+            self._pool_manager.create(self.pid, self._pool_idle_ttl,
+                                      self._pool_max_size, self._ioloop.time)
 
     @property
     def connection(self):
@@ -228,38 +230,16 @@ class TornadoSession(session.Session):
 
     def validate(self):
         """Validate the session can connect or has open connections to
-        PostgreSQL
+        PostgreSQL. As of ``1.10.3``
+
+        .. deprecated:: 1.10.3
+           As of 1.10.3, this method only warns about Deprecation
 
         :rtype: bool
 
         """
-        future = concurrent.TracebackFuture()
-
-        def on_connected(cf):
-            if cf.exception():
-                future.set_exception(cf.exception())
-                return
-
-            connection = cf.result()
-            fd = connection.fileno()
-
-            # The connection would have been added to the pool manager, free it
-            self._pool_manager.free(self.pid, connection)
-            self._ioloop.remove_handler(fd)
-
-            if fd in self._connections:
-                del self._connections[fd]
-            if fd in self._futures:
-                del self._futures[fd]
-
-            # Return the success in validating the connection
-            future.set_result(True)
-
-        # Grab a connection to PostgreSQL
-        self._ioloop.add_future(self._connect(), on_connected)
-
-        # Return the future for the query result
-        return future
+        warnings.warn(
+            'All functionality removed from this method', DeprecationWarning)
 
     def _connect(self):
         """Connect to PostgreSQL, either by reusing a connection from the pool
@@ -269,7 +249,7 @@ class TornadoSession(session.Session):
         :raises: pool.NoIdleConnectionsError
 
         """
-        future = concurrent.TracebackFuture()
+        future = concurrent.Future()
 
         # Attempt to get a cached connection from the connection pool
         try:
@@ -292,10 +272,16 @@ class TornadoSession(session.Session):
         :param tornado.concurrent.Future future: future for new conn result
 
         """
+        LOGGER.debug('Creating a new connection for %s', self.pid)
+
         # Create a new PostgreSQL connection
         kwargs = utils.uri_to_kwargs(self._uri)
 
-        connection = self._psycopg2_connect(kwargs)
+        try:
+            connection = self._psycopg2_connect(kwargs)
+        except (psycopg2.Error, OSError, socket.error) as error:
+            future.set_exception(error)
+            return
 
         # Add the connection for use in _poll_connection
         fd = connection.fileno()
@@ -309,19 +295,28 @@ class TornadoSession(session.Session):
 
             """
             if cf.exception():
+                self._cleanup_fd(fd, True)
                 future.set_exception(cf.exception())
 
             else:
 
-                # Add the connection to the pool
-                self._pool_manager.add(self.pid, connection)
+                try:
+                    # Add the connection to the pool
+                    LOGGER.debug('Connection established for %s', self.pid)
+                    self._pool_manager.add(self.pid, connection)
+                except (ValueError, pool.PoolException) as err:
+                    LOGGER.exception('Failed to add %r to the pool', self.pid)
+                    self._cleanup_fd(fd)
+                    future.set_exception(err)
+                    return
+
                 self._pool_manager.lock(self.pid, connection, self)
 
-                # Added in because psycopg2ct connects and leaves the
+                # Added in because psycopg2cffi connects and leaves the
                 # connection in a weird state: consts.STATUS_DATESTYLE,
                 # returning from Connection._setup without setting the state
                 # as const.STATUS_OK
-                if PYPY:
+                if utils.PYPY:
                     connection.status = extensions.STATUS_READY
 
                 # Register the custom data types
@@ -332,7 +327,7 @@ class TornadoSession(session.Session):
                 future.set_result(connection)
 
         # Add a future that fires once connected
-        self._futures[fd] = concurrent.TracebackFuture()
+        self._futures[fd] = concurrent.Future()
         self._ioloop.add_future(self._futures[fd], on_connected)
 
         # Add the connection to the IOLoop
@@ -362,7 +357,7 @@ class TornadoSession(session.Session):
         :raises: queries.ProgrammingError
 
         """
-        future = concurrent.TracebackFuture()
+        future = concurrent.Future()
 
         def on_connected(cf):
             """Invoked by the future returned by self._connect"""
@@ -377,18 +372,18 @@ class TornadoSession(session.Session):
             def completed(qf):
                 """Invoked by the IOLoop when the future has completed"""
                 if qf.exception():
-                    error = qf.exception()
-                    LOGGER.debug('Cleaning cursor due to exception: %r', error)
+                    self._incr_exceptions(conn)
+                    err = qf.exception()
+                    LOGGER.debug('Cleaning cursor due to exception: %r', err)
                     self._exec_cleanup(cursor, conn.fileno())
-                    future.set_exception(error)
+                    future.set_exception(err)
                 else:
-                    results = Results(cursor,
-                                      self._exec_cleanup,
-                                      conn.fileno())
-                    future.set_result(results)
+                    self._incr_executions(conn)
+                    value = Results(cursor, self._exec_cleanup, conn.fileno())
+                    future.set_result(value)
 
             # Setup a callback to wait on the query result
-            self._futures[conn.fileno()] = concurrent.TracebackFuture()
+            self._futures[conn.fileno()] = concurrent.Future()
 
             # Add the future to the IOLoop
             self._ioloop.add_future(self._futures[conn.fileno()],
@@ -396,7 +391,13 @@ class TornadoSession(session.Session):
 
             # Get the cursor, execute the query
             func = getattr(cursor, method)
-            func(query, parameters)
+            try:
+                func(query, parameters)
+            except Exception as error:
+                future.set_exception(error)
+
+        # Ensure the pool exists for the connection
+        self._ensure_pool_exists()
 
         # Grab a connection to PostgreSQL
         self._ioloop.add_future(self._connect(), on_connected)
@@ -413,31 +414,68 @@ class TornadoSession(session.Session):
 
         """
         LOGGER.debug('Closing cursor and cleaning %s', fd)
-        cursor.close()
-        self._pool_manager.free(self.pid, self._connections[fd])
-        self._ioloop.remove_handler(fd)
+        try:
+            cursor.close()
+        except (psycopg2.Error, psycopg2.Warning) as error:
+            LOGGER.debug('Error closing the cursor: %s', error)
 
+        self._cleanup_fd(fd)
+
+        # If the cleanup callback exists, remove it
+        if self._cleanup_callback:
+            self._ioloop.remove_timeout(self._cleanup_callback)
+
+        # Create a new cleanup callback to clean the pool of idle connections
+        self._cleanup_callback = self._ioloop.add_timeout(
+            self._ioloop.time() + self._pool_idle_ttl + 1,
+            self._pool_manager.clean, self.pid)
+
+    def _cleanup_fd(self, fd, close=False):
+        """Ensure the socket socket is removed from the IOLoop, the
+        connection stack, and futures stack.
+
+        :param int fd: The fd # to cleanup
+
+        """
+        self._ioloop.remove_handler(fd)
         if fd in self._connections:
+            try:
+                self._pool_manager.free(self.pid, self._connections[fd])
+            except pool.ConnectionNotFoundError:
+                pass
+            if close:
+                self._connections[fd].close()
             del self._connections[fd]
         if fd in self._futures:
             del self._futures[fd]
 
-    def _on_io_events(self, fd=None, events=None):
+    def _incr_exceptions(self, conn):
+        """Increment the number of exceptions for the current connection.
+
+        :param psycopg2.extensions.connection conn: the psycopg2 connection
+
+        """
+        self._pool_manager.get_connection(self.pid, conn).exceptions += 1
+
+    def _incr_executions(self, conn):
+        """Increment the number of executions for the current connection.
+
+        :param psycopg2.extensions.connection conn: the psycopg2 connection
+
+        """
+        self._pool_manager.get_connection(self.pid, conn).executions += 1
+
+    def _on_io_events(self, fd=None, _events=None):
         """Invoked by Tornado's IOLoop when there are events for the fd
 
         :param int fd: The file descriptor for the event
-        :param int events: The events raised
+        :param int _events: The events raised
 
         """
         if fd not in self._connections:
             LOGGER.warning('Received IO event for non-existing connection')
             return
-        try:
-            self._poll_connection(fd)
-        except OSError as error:
-            self._futures[fd].set_exception(
-                psycopg2.OperationalError('Connection error (%s)' % error)
-            )
+        self._poll_connection(fd)
 
     def _poll_connection(self, fd):
         """Check with psycopg2 to see what action to take. If the state is
@@ -448,8 +486,15 @@ class TornadoSession(session.Session):
         """
         try:
             state = self._connections[fd].poll()
+        except (OSError, socket.error) as error:
+            self._ioloop.remove_handler(fd)
+            if fd in self._futures and not self._futures[fd].done():
+                self._futures[fd].set_exception(
+                    psycopg2.OperationalError('Connection error (%s)' % error)
+                )
         except (psycopg2.Error, psycopg2.Warning) as error:
-            self._futures[fd].set_exception(error)
+            if fd in self._futures and not self._futures[fd].done():
+                self._futures[fd].set_exception(error)
         else:
             if state == extensions.POLL_OK:
                 if fd in self._futures and not self._futures[fd].done():
@@ -460,7 +505,9 @@ class TornadoSession(session.Session):
                 self._ioloop.update_handler(fd, ioloop.IOLoop.READ)
             elif state == extensions.POLL_ERROR:
                 self._ioloop.remove_handler(fd)
-                self._futures[fd].set_exception(psycopg2.Error('Poll Error'))
+                if fd in self._futures and not self._futures[fd].done():
+                    self._futures[fd].set_exception(
+                        psycopg2.Error('Poll Error'))
 
     def _psycopg2_connect(self, kwargs):
         """Return a psycopg2 connection for the specified kwargs. Extend for
